import os
import re
import shutil
from urllib import parse

from jirafs import exceptions, utils
from jirafs.plugin import CommandPlugin
from jirafs.ticketfolder import TicketFolder


class Command(CommandPlugin):
    """ Clone a new ticketfolder for the specified ticket URL"""

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"
    AUTOMATICALLY_INSTANTIATE_FOLDER = False

    TICKET_RE = re.compile(r".*\/browse\/(\w+-\d+)\/?")

    def handle(self, args, jira, path, **kwargs):
        ticket_url = args.ticket_url[0]

        ticket_url_parts = parse.urlparse(ticket_url)
        if not ticket_url_parts.netloc:
            default_server = utils.get_default_jira_server()
            ticket_url = parse.urljoin(default_server, "browse/" + ticket_url + "/")

        path = args.path[0] if args.path else None

        return self.cmd(path, ticket_url, jira)

    def clone_from_issue(self, match, ticket_url, path, jira):
        if not path:
            path = match.group(1)
        path = os.path.realpath(path)
        os.mkdir(path)

        try:
            folder = TicketFolder.initialize_ticket_folder(ticket_url, path, jira)

            utils.run_command_method_with_kwargs("pull", folder=folder)
        except BaseException:
            shutil.rmtree(path)
            raise

        folder.log(
            "Issue %s cloned successfully to %s", (folder.issue_url, folder.path,)
        )

        return folder

    def main(self, path, url, jira):
        match = self.TICKET_RE.match(url)
        if not match:
            raise exceptions.JirafsError("'%s' is not a valid Jira ticket URL." % url)

        return self.clone_from_issue(match, url, path, jira,)

    def add_arguments(self, parser):
        parser.add_argument("ticket_url", nargs=1, type=str)
        parser.add_argument(
            "path", nargs="*", type=str,
        )



import os
from unittest import TestCase, SkipTest
import shutil
import subprocess
import tempfile

from jira.client import JIRA

from jirafs.utils import run_command_method_with_kwargs
from jirafs.ticketfolder import TicketFolder


class IntegrationTestBase(TestCase):
    def setUp(self):
        super(IntegrationTestBase, self).setUp()

        self.jira_configured = True

        jira_env_settings = {
            "username": "INTEGRATION_TESTING_USERNAME",
            "url": "INTEGRATION_TESTING_URL",
            "project": "INTEGRATION_TESTING_PROJECT",
            "password": "INTEGRATION_TESTING_PASSWORD",
        }
        self.jira_env = {}
        for key, env_var in jira_env_settings.items():
            try:
                self.jira_env[key] = os.environ[env_var]
            except KeyError:
                raise SkipTest(
                    "Integration tests require the following environment "
                    "variables to be set: %s"
                    % (", ".join([v for k, v in jira_env_settings.items()]))
                )

        self.jira = JIRA(
            {"server": self.jira_env["url"]},
            basic_auth=(self.jira_env["username"], self.jira_env["password"],),
        )

        self.created_issues = []
        self.path = tempfile.mkdtemp()

    def get_jira(self, *args, **kwargs):
        return self.jira

    def get_ticket_folder_for_path(self, path):
        return TicketFolder(path, self.get_jira,)

    def _get_default_fields(self):
        return {
            "project": {"key": self.jira_env["project"],},
            "issuetype": {"name": "Task"},
        }

    def tearDown(self):
        for issue in self.created_issues:
            issue.delete()

        shutil.rmtree(self.path, ignore_errors=True)

    def create_issue(self, fields, *args, **kwargs):
        issue_fields = self._get_default_fields()
        issue_fields.update(fields)

        issue = self.jira.create_issue(issue_fields, *args, **kwargs)
        self.created_issues.append(issue)

        return issue

    def run_command(self, name, **kwargs):
        return run_command_method_with_kwargs(name, **kwargs)

    def run_from_shell(self, *args, **kwargs):
        if "path" not in kwargs:
            path = self.path
        else:
            path = kwargs["path"]

        return subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=path,
        )



import argparse
import codecs
import copy
import logging
import logging.config
import os
import shlex
import subprocess
import sys
import time
import traceback

from blessings import Terminal

try:
    from jira.utils import JIRAError
except ImportError:
    from jira.exceptions import JIRAError
from distutils.version import LooseVersion

from . import utils
from .exceptions import (
    GitCommandError,
    JiraInteractionFailed,
    JirafsError,
    NotTicketFolderException,
    UnknownMacroError,
    MacroError,
)


# Write data to stdout as UTF-8 bytes when there's no encoding specified
if sys.version_info < (3,) and sys.stdout.encoding is None:
    sys.stdout = codecs.getwriter("utf-8")(sys.stdout)


logger = logging.getLogger(__name__)


LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "pretty": {"format": "[%(levelname)s] %(message)s"},
        "file": {"format": "%(asctime)s\t%(levelname)s\t%(module)s\t%(message)s"},
    },
    "handlers": {
        "console": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "pretty",
            "stream": "ext://sys.stdout",
        }
    },
    "root": {"handlers": ["console"], "level": "INFO"},
}


def main():
    term = Terminal()
    if sys.version_info < (2, 7):
        raise RuntimeError(
            "Jirafs requires minimally version 2.7 of Python 2, or "
            "any version of Python 3.  Please upgrade your version of "
            "python before using Jirafs."
        )
    if utils.get_git_version() < LooseVersion("1.8"):
        raise RuntimeError(
            "Jirafs requires minimally version 1.8 of Git.  Please "
            "upgrade your version of git before using Jirafs."
        )

    commands = utils.get_installed_commands()

    parser = argparse.ArgumentParser(
        description="Edit Jira issues locally from your filesystem", add_help=False,
    )
    parser.add_argument("command", type=str, choices=commands.keys())
    parser.add_argument("--subtasks", action="store_true", default=False)
    parser.add_argument(
        "--log-level", default="INFO", dest="log_level",
    )
    parser.add_argument("--folder", default=os.getcwd())
    parser.add_argument(
        "--no-subfolders", action="store_true", default=False,
    )
    parser.add_argument(
        "--traceback", action="store_true", default=False,
    )
    parser.add_argument(
        "--debugger-port",
        default=58024,
        help=("(Requires --debugger) start debugger on this port."),
    )
    parser.add_argument(
        "--debugger",
        action="store_true",
        default=False,
        help=("Launch ptvsd debugger on --debugger-port."),
    )
    args, extra = parser.parse_known_args()

    if args.debugger:
        try:
            import ptvsd

            ptvsd.enable_attach(address=("localhost", args.debugger_port))
            print(
                f"{term.magenta}Waiting for debugger connection on "
                f"port {args.debugger_port}...{term.normal}"
            )
            ptvsd.wait_for_attach()
        except ImportError:
            print(f"{term.red}Module 'ptvsd' required for debugging." f"{term.normal}")
            sys.exit(1)

    logging.config.dictConfig(LOGGING)
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.getLevelName(args.log_level))

    command_name = args.command
    cmd_class = commands[command_name]

    # Subtasks
    if args.subtasks:
        cmd_class.RUN_FOR_SUBTASKS = True

    started = time.time()
    logger.debug("Command %s(%s) started", command_name, extra)
    jira = utils.lazy_get_jira()
    try:
        value = cmd_class.execute_command(
            extra, jira=jira, path=args.folder, command_name=command_name
        )
        logger.debug(
            "Command %s(%s) finished in %s seconds",
            command_name,
            extra,
            (time.time() - started),
        )
        if value:
            value.echo()
        sys.exit(value.return_code)
    except GitCommandError as e:
        print(
            "{t.red}Error (code: {code}) while running git "
            "command.{t.normal}".format(t=term, code=e.returncode)
        )
        print("")
        print("{t.red}Command:{t.normal}{t.red}{t.bold}".format(t=term))
        print("    {cmd}".format(cmd=e.command))
        print("{t.normal}".format(t=term))
        print("{t.red}Output:{t.normal}{t.red}{t.bold}".format(t=term))
        for line in e.output.decode("utf8").split("\n"):
            print("    %s" % line)
        print("{t.normal}".format(t=term))
        if args.traceback:
            traceback.print_exc()
        sys.exit(10)
    except NotTicketFolderException:
        if not getattr(cmd_class, "TRY_SUBFOLDERS", False):
            print(
                "{t.red}The command '{cmd}' must be ran from "
                "within an issue folder.{t.normal}".format(t=term, cmd=command_name)
            )
            sys.exit(20)
        elif args.no_subfolders:
            sys.exit(20)

        count_runs = 0
        for folder in os.listdir(os.getcwd()):
            full_path = os.path.join(os.getcwd(), folder,)
            if not os.path.isdir(full_path):
                continue

            try:
                full_args = copy.copy(sys.argv)
                if "--no-subfolders" not in full_args:
                    full_args.append("--no-subfolders")
                result = subprocess.call(
                    " ".join([shlex.quote(a) for a in full_args]),
                    cwd=full_path,
                    shell=True,
                )
                if result == 0:
                    count_runs += 1
            except NotTicketFolderException:
                pass
        if count_runs == 0:
            if args.traceback:
                traceback.print_exc()
            sys.exit(21)
    except UnknownMacroError as e:
        print(
            "{t.red}Jirafs encountered an unknown macro while processing "
            "your content: {t.normal}{t.red}{t.bold}{error}"
            "{t.normal}".format(t=term, error=str(e))
        )
        if args.traceback:
            traceback.print_exc()
        sys.exit(30)
    except MacroError as e:
        print(
            "{t.red}Jirafs encountered an error while processing a "
            "{t.bold}{macro}{t.normal}{t.red} macro: "
            "{t.normal}{t.red}{t.bold}{error}{t.normal}".format(
                t=term, macro=e.macro_name or "?", error=str(e)
            )
        )
        if args.traceback:
            traceback.print_exc()
        sys.exit(40)
    except JIRAError as e:
        print(
            "{t.red}Jirafs encountered an error while interacting with "
            "your Jira instance: {t.normal}{t.red}{t.bold}{error}"
            "{t.normal}".format(t=term, error=str(e))
        )
        if args.traceback:
            traceback.print_exc()
        sys.exit(70)
    except JiraInteractionFailed as e:
        print(
            "{t.red}Jira was unable to satisfy your "
            "request: {t.normal}{t.red}{t.bold}{error}{t.normal}".format(
                t=term, error=str(e)
            )
        )
        if args.traceback:
            traceback.print_exc()
        sys.exit(80)
    except JirafsError as e:
        print(
            "{t.red}Jirafs encountered an error processing your "
            "request: {t.normal}{t.red}{t.bold}{error}{t.normal}".format(
                t=term, error=str(e)
            )
        )
        if args.traceback:
            traceback.print_exc()
        sys.exit(90)



import io

from jirafs import constants, exceptions, utils
from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """ Push locally-committed changes to Jira """

    TRY_SUBFOLDERS = True
    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def get_valid_issue_link_types(self, folder):
        if not hasattr(self, "_valid_issue_link_types"):
            data = {}
            for item in folder.jira.issue_link_types():
                data[item.outward.lower()] = ("outward", item)
                data[item.inward.lower()] = ("inward", item)

            self._valid_issue_link_types = data
        return self._valid_issue_link_types

    def validate_issue(self, folder):
        status = folder.status()

        # Validate issue statuses
        valid_types = self.get_valid_issue_link_types(folder)
        links = status["ready"]["links"].get("issue", {})
        for target, data in links.items():
            if data[1] is None:
                continue
            if data[1].get("status") not in valid_types:
                raise exceptions.IssueValidationError(
                    u"{status} is not a valid issue link type for {target}; "
                    u"options include the following: {options}".format(
                        status=data[1].get("status"),
                        target=target,
                        options=", ".join(valid_types.keys()),
                    )
                )

    def main(self, folder, **kwargs):
        self.validate_issue(folder)
        with utils.stash_local_changes(folder):
            status = folder.status()

            if not folder.is_up_to_date():
                raise exceptions.LocalCopyOutOfDate(
                    "Your local copy is out-of-date.  You must use "
                    "the 'merge' command to update your local copy "
                    "before pushing changes."
                )

            file_meta = folder.get_remote_file_metadata(shadow=False)

            deleted = set()
            for filename in status["ready"]["deleted"]:
                folder.log(
                    'Deleting file "%s"', (filename,),
                )
                for attachment in folder.issue.fields.attachment:
                    if attachment.filename == filename and attachment.id not in deleted:
                        folder.jira.delete_attachment(attachment.id)
                        deleted.add(id)

            for filename in status["ready"]["files"]:
                upload = io.BytesIO(
                    folder.get_local_file_at_revision(filename, "HEAD", binary=True)
                )
                folder.log(
                    'Uploading file "%s"', (filename,),
                )
                # Delete the existing issue if there is one
                for attachment in folder.issue.fields.attachment:
                    if attachment.filename == filename:
                        attachment.delete()
                upload.seek(0)
                attachment = folder.jira.add_attachment(
                    folder.ticket_number, upload, filename=filename,
                )
                file_meta[filename] = attachment.created

            folder.set_remote_file_metadata(file_meta, shadow=False)

            comment = folder.get_new_comment(clear=True, ready=True)
            if comment:
                folder.log(
                    u'Adding comment "{comment}"'.format(
                        comment=self.truncate_field_value(comment)
                    )
                )
                folder.jira.add_comment(folder.ticket_number, comment)

            collected_updates = {}
            for field, diff_values in status["ready"]["fields"].items():
                collected_updates[field] = diff_values[1]

            if collected_updates:
                folder.log('Updating fields "%s"', (collected_updates,))
                folder.issue.update(**collected_updates)

            links = status["ready"]["links"]
            statuses = self.get_valid_issue_link_types(folder)
            for target, data in links.get("issue", {}).items():
                orig = data[0]
                new = data[1]
                other = folder.jira.issue(target)
                if orig is None:
                    # New links
                    status_data = statuses[new["status"]]
                    args = [
                        status_data[1].name,
                    ]
                    if status_data[0] == "inward":
                        args.extend([other, folder.issue])
                    elif status_data[0] == "outward":
                        args.extend([folder.issue, other])
                    folder.jira.create_issue_link(*args)
                elif new is None:
                    # Deleted links
                    for existing_link in folder.issue.fields.issuelinks:
                        if (
                            hasattr(existing_link, "inwardIssue")
                            and existing_link.inwardIssue.key == target
                        ):
                            existing_link.delete()
                        if (
                            hasattr(existing_link, "outwardIssue")
                            and existing_link.outwardIssue.key == target
                        ):
                            existing_link.delete()
                else:
                    # Changed links
                    for existing_link in folder.issue.fields.issuelinks:
                        if (
                            hasattr(existing_link, "inwardIssue")
                            and existing_link.inwardIssue.key == target
                        ):
                            existing_link.type = statuses[new["status"]][1]
                            existing_link.update()
                        if (
                            hasattr(existing_link, "outwardIssue")
                            and existing_link.outwardIssue.key == target
                        ):
                            existing_link.type = statuses[new["status"]][1]
                            existing_link.update()

            links = status["ready"]["links"]
            remote_links = folder.jira.remote_links(folder.issue)
            # Workaround for bug in python-jira:
            folder.jira._applicationlinks = []
            for target, data in links.get("remote", {}).items():
                orig = data[0]
                new = data[1]
                if orig is None:
                    # New links
                    link_object = {
                        "url": target,
                        "title": new["description"],
                    }
                    folder.jira.add_remote_link(folder.issue, link_object)
                elif new is None:
                    # Deleted links
                    for existing_link in remote_links:
                        if existing_link.object.url == target:
                            existing_link.delete()
                else:
                    # Changed links
                    for existing_link in remote_links:
                        if existing_link.object.url == target:
                            existing_link.update(
                                {"url": target, "title": new["description"]}
                            )

            # Commit local copy
            folder.run_git_command("reset", "--soft", failure_ok=True)
            folder.run_git_command(
                "add", folder.get_path(".jirafs/remote_files.json"), failure_ok=True
            )
            folder.run_git_command(
                "add", folder.get_path(constants.TICKET_NEW_COMMENT), failure_ok=True
            )
            folder.run_git_command(
                "commit", "-m", "Pushed local changes", failure_ok=True
            )

            # Commit changes to remote copy, too, so we record remote
            # file metadata.
            folder.run_git_command("fetch", shadow=True)
            folder.run_git_command("merge", "origin/master", shadow=True)
            folder.run_git_command("add", "-A", shadow=True)
            folder.run_git_command(
                "commit", "-m", "Pulled remote changes", failure_ok=True, shadow=True
            )
            folder.run_git_command("push", "origin", "jira", shadow=True)
            pull_result = run_command_method_with_kwargs("pull", folder=folder)
            return pull_result[1]



from jirafs import utils
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """Create a subtask of a given issue."""

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def main(self, folder, args, **kwargs):
        summary = " ".join(args.summary)

        folder.jira.create_issue(
            fields={
                "project": {"key": folder.issue.fields.project.key},
                "summary": summary,
                "issuetype": {"name": "Sub-task"},
                "parent": {"id": folder.issue.key},
            }
        )

        commands = utils.get_installed_commands()
        jira = utils.lazy_get_jira()
        commands["fetch"].execute_command(
            [], jira=jira, path=folder.path, command_name="fetch",
        )

    def add_arguments(self, parser):
        parser.add_argument("summary", nargs="+")



import configparser

from jirafs import utils
from jirafs.exceptions import NotTicketFolderException
from jirafs.plugin import CommandResult, DirectOutputCommandPlugin
from jirafs.ticketfolder import TicketFolder


class Command(DirectOutputCommandPlugin):
    """ Get, set, or list global or per-folder configuration values """

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"
    AUTOMATICALLY_INSTANTIATE_FOLDER = False

    def main(self, args, jira, path, parser, **kwargs):
        if args.global_config:
            config = utils.get_config()
        else:
            try:
                folder = TicketFolder(path, jira)
                config = folder.get_config()
            except NotTicketFolderException:
                config = utils.get_config()

        return_value = None
        if args.list:
            if len(args.params) != 0:
                parser.error("--list requires no parameters.")
            return_value = self.list(config)
        elif args.get:
            if len(args.params) != 1:
                parser.error(
                    "--get requires exactly one parameter, the configuration "
                    "value to display."
                )
            section, key = self.get_section_and_key(args.params[0])
            return_value = self.get(config, section, key)
        elif args.set:
            if len(args.params) != 2:
                parser.error(
                    "--set requires exactly two parameters, the configuration "
                    "key, and the configuration value."
                )
            section, key = self.get_section_and_key(args.params[0])
            value = args.params[1]

            if args.global_config:
                return_value = self.set_global(section, key, value)
            else:
                try:
                    folder = TicketFolder(path, jira)
                    return_value = self.set_local(folder, section, key, value)
                except NotTicketFolderException:
                    parser.error(
                        "Not currently within a ticket folder.  To set a "
                        "global configuration value, use the --global option."
                    )

        return return_value

    def get_section_and_key(self, string):
        return string.rsplit(".", 1)

    def set_global(self, section, key, value):
        return utils.set_global_config_value(section, key, value)

    def set_local(self, folder, section, key, value):
        return folder.set_config_value(section, key, value)

    def get(self, config, section, key):
        try:
            value = config.get(section, key)
            return CommandResult(value)
        except configparser.Error:
            pass

    def list(self, config):
        lines = CommandResult()

        for section in config.sections():
            parameters = config.items(section)
            for key, value in parameters:
                lines = lines.add_line(
                    u"{section}.{key}={value}", section=section, key=key, value=value
                )

        return lines

    def add_arguments(self, parser):
        parser.add_argument("--list", action="store_true")
        parser.add_argument("--get", action="store_true")
        parser.add_argument("--set", action="store_true")
        parser.add_argument(
            "--global", dest="global_config", default=False, action="store_true"
        )
        parser.add_argument("params", nargs="*")

    def parse_arguments(self, parser, args):
        args = super(Command, self).parse_arguments(parser, args)

        if not args.list and not args.get and not args.set:
            args.list = True

        return args



import io
import os

from .exceptions import GitCommandError


class GitRevisionReader(object):
    def __init__(self, folder, revision):
        self.folder = folder
        self.revision = revision
        super(GitRevisionReader, self).__init__()

    def get_file_contents(self, path):
        try:
            return self.folder.get_local_file_at_revision(
                path, self.revision, failure_ok=False
            )
        except GitCommandError:
            return ""


class WorkingCopyReader(object):
    def __init__(self, folder, path):
        self.folder = folder
        self.path = path
        super(WorkingCopyReader, self).__init__()

    def get_file_contents(self, path):
        full_path = os.path.join(self.path, path)

        with io.open(
            self.folder.get_local_path(full_path), "r", encoding="utf-8"
        ) as _in:
            return _in.read().strip()



import json

from jirafs.plugin import CommandPlugin, CommandResult


class Command(CommandPlugin):
    """ Get the status of the current ticketfolder """

    TRY_SUBFOLDERS = True
    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def handle(self, args, folder, **kwargs):
        return self.cmd(folder, args.format)

    def add_arguments(self, parser):
        parser.add_argument("--format", default="text", choices=["text", "json"])

    def main(self, folder, output_format="text"):
        return folder.status()

    def cmd(self, folder, output_format="text"):
        status = self.main(folder)
        if output_format == "json":
            return self.status_json(folder, status)
        return self.status_text(folder, status)

    def status_json(self, folder, status):
        result = CommandResult()
        result = result.add_line(json.dumps(status), no_format=True)
        return result

    def has_changes(self, section, *keys):
        if not keys:
            keys = ["files", "deleted", "fields", "new_comment", "links"]
        for key in keys:
            if key in section and section[key]:
                return True

    def status_text(self, folder, folder_status):
        result = CommandResult()

        result = result.add_line(
            "On ticket {ticket} ({url})",
            ticket=folder.ticket_number,
            url=folder.cached_issue.permalink(),
        )
        if not folder_status["up_to_date"]:
            result = result.add_line(
                "{t.magenta}Warning: unmerged upstream changes exist; "
                "run `jirafs merge` to merge them into your local copy."
                "{t.normal}"
            )

        printed_changes = False
        ready = folder_status["ready"]
        if self.has_changes(ready):
            printed_changes = True
            result = result.add_line("")
            result = result.add_line(
                "Ready for upload; use `jirafs push` to update Jira"
            )
            result = self.format_field_changes(
                ready, "green", result=result, files_post_message="upload attachment"
            )

        staged = folder_status["uncommitted"]
        if self.has_changes(staged):
            printed_changes = True
            result = result.add_line("")
            result = result.add_line(
                "Uncommitted changes: use `jirafs submit` to submit changes "
                "to Jira, or use `jirafs commit` to commit your changes "
                "for submission during a later `jirafs push`."
            )
            result = self.format_field_changes(
                staged, "red", result=result, files_post_message="upload attachment"
            )

        local_uncommitted = folder_status["local_uncommitted"]
        if self.has_changes(local_uncommitted, "files"):
            printed_changes = True
            result = result.add_line("")
            result = result.add_line(
                "Uncommitted changes prevented from being sent to Jira "
                "because they match at least one of the patterns in your "
                "`.jirafs_local` file; use `jirafs commit` to track these "
                "changes."
            )
            result = result.add_line(
                "Note: these files will {t.bold}not{t.normal} "
                "be uploaded to Jira even after being committed."
            )
            result = self.format_field_changes(
                local_uncommitted,
                "cyan",
                files_post_message="track in repository",
                result=result,
            )

        conflicts = folder_status["conflicts"]
        if self.has_changes(conflicts, "files"):
            printed_changes = True
            result = result.add_line("")
            result = result.add_line(
                "Local changes conflicted with changes merged "
                "from your remote Jira issue.  Please correct "
                "the following files and then run `jirafs commit`:"
            )
            result = self.format_field_changes(
                conflicts, "yellow", files_post_message="conflicted", result=result,
            )

        if not printed_changes:
            result = result.add_line("No changes found")
        else:
            result = result.add_line("")
            result = result.add_line(
                "For more detail about these changes, run `jirafs diff`"
            )

        return result

    def format_field_changes(self, changes, color, files_post_message="", result=None):
        if result is None:
            result = CommandResult()

        for filename in changes.get("files", []):
            result = result.add_line(
                "\t{t.%s}{filename}{t.normal}{post_message}" % color,
                filename=filename,
                post_message=(f" ({files_post_message})" if files_post_message else ""),
            )
        for filename in changes.get("deleted", []):
            result = result.add_line(
                "\t{t.%s}{filename}{t.normal} {post_message}" % color,
                filename=filename,
                post_message="(deleted)",
            )
        for link, data in changes.get("links", {}).get("remote", {}).items():
            orig = data[0]
            new = data[1]
            if new is not None:
                if new.get("description"):
                    description = new["description"]
                else:
                    description = "(Untitled)"

                result = result.add_line(
                    "\t{t.%s}{description}: "
                    "{link}{t.normal} {post_message}" % (color,),
                    description=description,
                    link=link,
                    post_message=(
                        " (changed remote link)" if orig else " (new remote link)"
                    ),
                )
            else:
                if orig.get("description"):
                    description = orig["description"]
                else:
                    description = "(Untitled)"

                result = result.add_line(
                    "\t{t.%s}{description}: "
                    "{link}{t.normal} {post_message}" % (color,),
                    description=description,
                    link=link,
                    post_message="(removed remote link)",
                )
        for link, data in changes.get("links", {}).get("issue", {}).items():
            orig = data[0]
            new = data[1]
            if new is not None:
                if new.get("status"):
                    status = new["status"]
                else:
                    status = "(Untitled)"
                result = result.add_line(
                    "\t{t.%s}{status}: {link}{t.normal} {post_message}" % (color),
                    status=status.title(),
                    link=link,
                    post_message=(
                        "(changed issue link)" if orig else "(new issue link)"
                    ),
                )
            else:
                if orig.get("status"):
                    status = orig["status"]
                else:
                    status = "(Untitled)"

                result = result.add_line(
                    "\t{t.%s}{status}: {link}{t.normal} {post_message}" % (color,),
                    status=status.title(),
                    link=link,
                    post_message="(removed issue link)",
                )
        for field, value_set in changes.get("fields", {}).items():
            result = result.add_line("\t{t.%s}{field}{t.normal}" % color, field=field,)
        if changes.get("new_comment", ""):
            result = result.add_line("\t{t.%s}[New Comment]{t.normal}" % color)

        return result



from jirafs import exceptions, utils
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Commit local changes for later submission to Jira """

    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def handle(self, args, folder, **kwargs):
        return self.cmd(folder, args.message, *args.git_arguments)

    def add_arguments(self, parser):
        parser.add_argument("-m", "--message", dest="message", default="Untitled")
        parser.add_argument("git_arguments", nargs="*")

    def main(self, folder, message, *args):
        # Run all macros to ensure that we've written any file
        # changes that such macros might perform
        folder.process_macros_for_all_fields()

        original_hash = folder.run_git_command("rev-parse", "master")
        folder.run_git_command("add", "-A")
        try:
            folder.run_git_command("commit", "-m", message, *args)
            final_hash = folder.run_git_command("rev-parse", "master")
            folder.log("Changes committed; current hash is %s" % final_hash)
        except exceptions.GitCommandError:
            print("Nothing to commit")
        final_hash = folder.run_git_command("rev-parse", "master")
        return utils.PostStatusResponse(original_hash == final_hash, final_hash)



import os

from .base import IntegrationTestBase

from jirafs import constants


class TestCloneIssue(IntegrationTestBase):
    def test_cloning_issue_basic(self):
        issue_path = os.path.join(self.path, "issue")
        data = {
            "description": "This is a test description",
            "summary": "This is a test summary",
        }

        issue = self.create_issue(data)

        self.run_command(
            "clone", path=issue_path, url=issue.permalink(), jira=self.get_jira
        )

        folder = self.get_ticket_folder_for_path(issue_path)
        fields = folder.get_fields()

        # Assert that the field data is on-disk as expected
        for key, value in data.items():
            self.assertEqual(value, fields.get(key))

    def test_cloning_issue_unicode_fields(self):
        issue_path = os.path.join(self.path, "issue")
        data = {
            "description": u"\u041f\u0440\u0438\u0432\u0435\u0442",
            "summary": u"\u0410\u043b\u043e",
        }

        issue = self.create_issue(data)

        self.run_command(
            "clone", path=issue_path, url=issue.permalink(), jira=self.get_jira
        )

        folder = self.get_ticket_folder_for_path(issue_path)
        fields = folder.get_fields()

        # Assert that the field data is on-disk as expected
        for key, value in data.items():
            self.assertEqual(value, fields.get(key))

    def test_cloning_issue_files_exist(self):
        issue_path = os.path.join(self.path, "issue")
        data = {
            "description": "This is a test description",
            "summary": "This is a test summary",
        }

        issue = self.create_issue(data)

        self.run_command(
            "clone", path=issue_path, url=issue.permalink(), jira=self.get_jira
        )

        folder = self.get_ticket_folder_for_path(issue_path)

        description_filename = constants.TICKET_FILE_FIELD_TEMPLATE.format(
            field_name="description"
        )
        expected_files = [
            constants.TICKET_DETAILS,
            constants.TICKET_COMMENTS,
            constants.TICKET_NEW_COMMENT,
            constants.TICKET_LINKS,
            description_filename,
        ]
        all_files = os.listdir(issue_path)

        for filename in expected_files:
            self.assertIn(filename, all_files)

        with open(folder.get_path(description_filename), "r") as handle:
            self.assertEqual(
                data["description"], handle.read().strip(),
            )

    def test_post_cloning_status(self):
        issue_path = os.path.join(self.path, "issue")
        data = {
            "description": "This is a test description",
            "summary": "This is a test summary",
        }

        issue = self.create_issue(data)

        self.run_command(
            "clone", path=issue_path, url=issue.permalink(), jira=self.get_jira
        )

        stdout = self.run_command(
            "status", method="cmd", folder=self.get_ticket_folder_for_path(issue_path)
        )

        assert_strings = [
            "On ticket %s" % issue.key,
            "browse/%s" % issue.key,
            "No changes found",
        ]
        for expectation in assert_strings:
            self.assertIn(
                expectation, stdout,
            )



from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """ Fetch and merge remote changes """

    RUN_FOR_SUBTASKS = True
    TRY_SUBFOLDERS = True
    MIN_VERSION = "2.0.0"
    MAX_VERSION = "3.0.0"

    def main(self, folder, **kwargs):
        fetch_result = run_command_method_with_kwargs("fetch", folder=folder)
        merge_result = run_command_method_with_kwargs("merge", folder=folder)

        return fetch_result, merge_result



import collections
import configparser
import contextlib
import getpass
import logging
import os
import pkg_resources
import re
import subprocess

from typing import Dict, Tuple

from jira.client import JIRA
from distutils.version import LooseVersion

from . import constants
from .plugin import CommandPlugin, Plugin


logger = logging.getLogger(__name__)


def convert_to_boolean(string):
    if string.upper().strip() in ["Y", "YES", "ON", "ENABLED", "ENABLE", "TRUE"]:
        return True
    elif string.upper().strip() in ["N", "NO", "OFF", "DISABLED", "DISABLE", "FALSE"]:
        return False
    return None


@contextlib.contextmanager
def stash_local_changes(repo):
    # Only v10 of repositories will properly handle stashing local
    # changes since the `version` file was previously untracked.
    try:
        if repo.version >= 10:
            repo.run_git_command("stash", failure_ok=True)
        yield
    finally:
        if repo.version >= 10:
            # This looks a little weird, I know, but what we're doing
            # is forcing a stash pop; even if it causes conflict markers,
            # that's better than just dropping the stash on the floor.
            repo.run_git_command("stash", "apply", failure_ok=True)
            repo.run_git_command("stash", "drop", failure_ok=True)


def get_user_input(message, options=None, boolean=False, password=False):
    if not constants.ALLOW_USER_INPUT:
        raise RuntimeError("User input is disabled")

    value = None
    while value is None:
        if not password:
            raw_value = input(message + " ")
        else:
            raw_value = getpass.getpass(message + " ")

        if options and raw_value not in options:
            print("'%s' is not a valid response" % raw_value)
            continue
        elif boolean:
            result = convert_to_boolean(raw_value)
            if result is not None:
                value = result
            else:
                print("'%s' is not a valid response" % raw_value)
        elif raw_value.strip():
            value = raw_value
        else:
            print("Please enter a response")

    return value


def run_command_method_with_kwargs(command, method=None, **kwargs):
    if method is None:
        method = "main"
    installed_commands = get_installed_commands()
    return getattr(installed_commands[command](), method)(**kwargs)


def get_installed_commands():
    possible_commands = {}
    for entry_point in pkg_resources.iter_entry_points(group="jirafs_commands"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, CommandPlugin):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `jirafs.plugin.CommandPlugin`.",
                entry_point,
            )
            continue
        possible_commands[entry_point.name] = loaded_class

    return possible_commands


def get_installed_plugins(subclass=Plugin):
    possible_plugins = {}
    for entry_point in pkg_resources.iter_entry_points(group="jirafs_plugins"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, Plugin):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `jirafs.plugin.Plugin`.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, subclass):
            # Do not emit a warning here -- this filter occurs
            # when we've limited the acceptable plugins to
            # a subset of jirafs.plugin.Plugin classes
            # intentionally, so this test failing is not
            # an error.
            continue
        possible_plugins[entry_point.name] = loaded_class

    return possible_plugins


def get_config_path(filename):
    if filename.startswith("/"):
        return filename
    return os.path.expanduser("~/%s" % filename)


def get_config(additional_configs=None, include_global=True):
    filenames = []
    if include_global:
        filenames.append(get_config_path(constants.GLOBAL_CONFIG))
    if additional_configs:
        filenames.extend(additional_configs)

    parser = configparser.RawConfigParser()
    parser.read(filenames)
    return parser


def set_global_config_value(section, key, value):
    config = get_config()
    if not config.has_section(section):
        config.add_section(section)
    config.set(section, key, value)
    with open(get_config_path(constants.GLOBAL_CONFIG), "w") as out:
        config.write(out)


def get_default_jira_server(config=None):
    if config is None:
        config = get_config()

    if not config.has_section(constants.CONFIG_JIRA):
        config.add_section(constants.CONFIG_JIRA)

    if not config.has_option(constants.CONFIG_JIRA, "server"):
        server_url_finder = re.compile(r"(?P<domain>.*)/browse/.*")
        value = get_user_input("Default Jira URL (or example issue URL): ")
        match = server_url_finder.match(value)
        if match:
            value = match.groupdict()["domain"]

        config.set(constants.CONFIG_JIRA, "server", value)

    with open(
        os.path.expanduser("~/%s" % constants.GLOBAL_CONFIG), "w"
    ) as global_config:
        config.write(global_config)

    return config.get(constants.CONFIG_JIRA, "server")


def get_jira(domain=None, config=None):
    def get_section(config, domain):
        if domain is None:
            return constants.CONFIG_JIRA

        valid_options = [
            domain.strip("/"),
            domain.strip("/") + "/",
        ]

        for option in valid_options:
            if config.has_section(option):
                return option

        config.add_section(valid_options[0])
        return valid_options[0]

    if config is None:
        config = get_config()

    if not config.has_section(constants.CONFIG_MAIN):
        config.add_section(constants.CONFIG_MAIN)

    ask_to_save = True
    if config.has_option(constants.CONFIG_MAIN, "ask_to_save"):
        ask_to_save = convert_to_boolean(
            config.get(constants.CONFIG_MAIN, "ask_to_save")
        )

    login_data = {"verify": True}

    section = get_section(config, domain)

    if domain is not None:
        login_data["server"] = domain
    else:
        login_data["server"] = get_default_jira_server(config)
        # Config may have been changed as a result of the above; reload.
        config = get_config()

    if not config.has_option(section, "username"):
        value = get_user_input("Jira Username (%s):" % login_data["server"])
        login_data["username"] = value
        config.set(section, "username", value)
    else:
        login_data["username"] = config.get(section, "username")

    if not config.has_option(section, "password"):
        value = get_user_input(
            "Jira Password (%s):" % login_data["server"], password=True,
        )
        login_data["password"] = value

        if ask_to_save:
            save = get_user_input("Save Jira Password (Y/N)?", boolean=True)
            if save:
                config.set(section, "password", value)
    else:
        login_data["password"] = config.get(section, "password")

    if config.has_option(section, "verify"):
        value = convert_to_boolean(config.get(section, "verify"))
        if value is None:
            # This means that we weren't able to convert the string's value
            # to a boolean, and it's probably instead a path to a PEM file.
            login_data["verify"] = config.get(section, "verify")
        else:
            login_data["verify"] = value

    # Prevent python-jira from checking to see if its out of date.
    login_data["check_update"] = False

    basic_auth = (
        login_data.pop("username"),
        login_data.pop("password"),
    )
    jira = JIRA(login_data, basic_auth=basic_auth)

    with open(get_config_path(constants.GLOBAL_CONFIG), "w") as global_config:
        config.write(global_config)

    return jira


def get_git_version():
    result = subprocess.check_output(
        ["git", "--version"], stderr=subprocess.PIPE,
    ).decode("utf8")
    version_string = re.match("git version ([0-9.]+).*", result).group(1)
    return LooseVersion(version_string)


def lazy_get_jira():
    return lambda domain, config=None: get_jira(domain, config)


def format_date(folder, date):
    date_format = constants.DEFAULT_DATE_FORMAT
    config = folder.get_config()
    if config.has_section(constants.CONFIG_MAIN):
        if config.has_option(constants.CONFIG_MAIN, "date_format"):
            date_format = config.get(constants.CONFIG_MAIN, "date_format")

    return date.strftime(date_format)


def find_files_referenced_in_markup(markup: str) -> Dict[str, Tuple[str, int, int]]:
    matched_files = {}
    finders = [
        re.compile(r"(!(?P<filename>\w.*)!)"),
        re.compile(r"(\[\^(?P<filename>\w.*)\])"),
    ]
    for finder in finders:
        for match in finder.finditer(markup):
            filename = match.groupdict()["filename"]
            if "|" in filename:
                filename = filename.split("|", 1)[0]

            matched_files[filename] = (
                match.group(0),
                match.start(0),
                match.end(0),
            )

    return matched_files


PostStatusResponse = collections.namedtuple("PostStatusResponse", ["new", "hash"])
